/*
 * Copyright (c) 2012 Imre Fazekas.
 *  All rights reserved.
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *
 *  Redistributions of source code must retain the above copyright notice, this
 *  list of conditions and the following disclaimer.
 *
 *  Redistributions in binary form must reproduce the above copyright notice,
 *  this list of conditions and the following disclaimer in the documentation
 *  and/or other materials provided with the distribution.
 *  Neither the name of the Brillien nor the names of its
 *  terms and concepts may be used to endorse or promote products derived from this
 *  software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

package com.vii.brillien.services.memcached;

import com.vii.brillien.core.component.db.CouchPresence;
import com.vii.brillien.core.management.BrillienServices;
import com.vii.brillien.ignition.BrillienContext;
import com.vii.brillien.ignition.BrillienStatics;
import com.vii.brillien.kernel.BrillienException;
import com.vii.brillien.kernel.annotations.PresenceService;
import com.vii.brillien.kernel.annotations.lifecycle.Resident;
import com.vii.streamline.services.error.StreamLineException;
import com.vii.streamline.services.json.JsonServices;
import com.vii.streamline.structures.collections.InnerList;
import net.spy.memcached.AddrUtil;
import net.spy.memcached.BinaryConnectionFactory;
import net.spy.memcached.MemcachedClient;
import org.jcouchdb.document.BaseDocument;

import java.io.IOException;
import java.util.List;

/**
 * Specialized CouchPresence providing caching solution via Memcached service. It stores any document possessing the Cachable interface in MCached too.
 */
@PresenceService(logLevel = PresenceService.FINE)
@Resident
public class CouchMPresence extends CouchPresence {

    /**
     *  MemcachedClient clients representing a connection pool to the Memcached server
     */
    protected MemcachedClient[]     m = null;

    /**
     * Tells whether the connection has been set up
     */
    protected boolean               connected;

    /**
     * Addresses of the Memcached server nodes
     */
    protected List<String>          mcachedAddresses;

    /**
     * Size of the connection pool
     */
    protected Long                  mcachedConnectionPoolSize;

    /**
     * Timout of the record to be stored
     */
    protected Long                  mcachedTimout;

    /**
     * Namespace to be used for every key associated to the documents to be stored
     */
    protected String                mcachedNameStapce;


    {
        connected                   = false;
        mcachedAddresses            = new InnerList<String>("");
        mcachedConnectionPoolSize   = 16l;
        mcachedTimout               = 60l * 60l;
        mcachedNameStapce           = BrillienContext.SERVER_NAME + BrillienServices.NAME_DELIMETER + BrillienStatics.SYSTEM_ID + BrillienServices.NAME_DELIMETER;
    }

    /**
     * Getter method for the field mcachedAddresses
     */
    public List<String> getMcachedAddresses() {
        return mcachedAddresses;
    }

    /**
     * Setter method for the field mcachedAddresses
     */
    public void setMcachedAddresses(List<String> mcachedAddresses) {
        this.mcachedAddresses = mcachedAddresses;
    }

    /**
     * Getter method for the field mcachedConnectionPoolSize
     */
    public Long getMcachedConnectionPoolSize() {
        return mcachedConnectionPoolSize;
    }

    /**
     * Setter method for the field mcachedConnectionPoolSize
     */
    public void setMcachedConnectionPoolSize(Long mcachedConnectionPoolSize) {
        this.mcachedConnectionPoolSize = mcachedConnectionPoolSize;
    }

    /**
     * Getter method for the field mcachedTimout
     */
    public Long getMcachedTimout() {
        return mcachedTimout;
    }

    /**
     * Setter method for the field mcachedTimout
     */
    public void setMcachedTimout(Long mcachedTimout) {
        this.mcachedTimout = mcachedTimout;
    }

    /**
     * Getter method for the field mcachedNameStapce
     */
    public String getMcachedNameStapce() {
        return mcachedNameStapce;
    }

    /**
     * Setter method for the field mcachedNameStapce
     */
    public void setMcachedNameStapce(String mcachedNameStapce) {
        this.mcachedNameStapce = mcachedNameStapce;
    }


    /**
     * Set up connection(s) to the Memcached node(s)
     */
    public void connect() throws BrillienException {
        m = new MemcachedClient[ mcachedConnectionPoolSize.intValue() ];
        try {
            for (int i = 0; i < mcachedConnectionPoolSize; i++) {
                MemcachedClient c = new MemcachedClient(
                        new BinaryConnectionFactory(),
                        AddrUtil.getAddresses( mcachedAddresses ));
                m[i] = c;
            }
            connected = true;
        } catch (IOException e) {
            throw new BrillienException(StreamLineException.ERROR_BROKEN_EXTERNAL, "Unable to establish connection to Memcached.", e);
        }
    }


    @Override
    public void plan() throws BrillienException {
        super.plan();

        try{
            connect();
        } catch (Exception e){
            throw new BrillienException( StreamLineException.ERROR_ILLEGAL_INNER_STATE, e );
        }
    }

    /**
     * Gets a random connection from the pool
     */
    public MemcachedClient getCache() {
        MemcachedClient c = null;
        try {
            int i = (int) (Math.random() * mcachedConnectionPoolSize);
            c = m[i];
        } catch (Exception e) {
            exceptionLog( CouchMPresence.class.getName(), "getCache", e );
        }
        return c;
    }

    /**
     * Retrieves a record from the cache identified by the given key
     */
    public Object get(String key) {
        Object o = getCache().get( mcachedNameStapce + key );
        if (o == null) {
            log("Cache MISS for KEY: " + key);
        } else {
            log("Cache HIT for KEY: " + key);
        }
        return o;
    }

    /**
     * Stores the given object for the given amount of time identified by the given key
     */
    public void set(String key, int ttl, final Object o) {
        getCache().set( mcachedNameStapce + key, ttl, o );
    }

    /**
     * Reomves an object form the cache identified by the given key
     */
    public Object delete(String key) {
        return getCache().delete( mcachedNameStapce + key );
    }


    /**
     * Extends the overriden method by storing the document in the cache too
     */
    @Override
    public boolean storeDocument( BaseDocument db) throws StreamLineException {
        boolean b = super.storeDocument( db );

        if( connected && db instanceof Cachable && ((Cachable)db).toBeCached() ){
            set( db.getId(), mcachedTimout.intValue(), JsonServices.toJSON( db ) );
        }

        return b;
    }

    /**
     * Extends the overriden method by updating the document in the cache too
     */
    @Override
    public boolean updateDocument(BaseDocument db) throws StreamLineException {
        boolean b = super.updateDocument(db);

        if( connected && db instanceof Cachable && ((Cachable)db).toBeCached() ){
            set( db.getId(), mcachedTimout.intValue(), JsonServices.toJSON( db ) );
        }

        return b;
    }

    /**
     * Extends the overriden method by removing the document from the cache too
     */
    @Override
    public boolean deleteDocument( BaseDocument db) throws StreamLineException{
        boolean b = super.deleteDocument(db);

        if( connected && db instanceof Cachable && ((Cachable)db).toBeCached() ){
            delete( db.getId() );
        }

        return b;
    }

    /**
     * Extends the overriden method by searchomg the document in the cache first
     */
    @Override
    public <T extends BaseDocument> T getDocumentById(Class<T> classReference, String id) throws StreamLineException {
        if( connected ){
            Object o = get( id );
            if( o != null ){
                try{
                    return JsonServices.<T>parseJSON( o.toString(), classReference );
                } catch (Exception e){ }
            }
        }

        return super.<T>getDocumentById( classReference, id );
    }

    @Override
    public String toString() {
        return "CouchMPresence{" +
                "mcachedAddresses=" + mcachedAddresses +
                ", mcachedConnectionPoolSize=" + mcachedConnectionPoolSize +
                ", mcachedTimout=" + mcachedTimout +
                ", mcachedNameStapce='" + mcachedNameStapce + '\'' +
                '}';
    }

}
