/*
 * Copyright (c) 2012 Imre Fazekas.
 *  All rights reserved.
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *
 *  Redistributions of source code must retain the above copyright notice, this
 *  list of conditions and the following disclaimer.
 *
 *  Redistributions in binary form must reproduce the above copyright notice,
 *  this list of conditions and the following disclaimer in the documentation
 *  and/or other materials provided with the distribution.
 *  Neither the name of the Brillien nor the names of its
 *  terms and concepts may be used to endorse or promote products derived from this
 *  software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

package com.vii.brillien.services.websocket.applications.purity;

import com.vii.brillien.core.management.BrillienServices;
import com.vii.brillien.ignition.BrillienContext;
import com.vii.brillien.ignition.transport.BrillienCommunication;
import com.vii.brillien.ignition.transport.BrillienMediator;
import com.vii.brillien.ignition.transport.LoaderComplyer;
import com.vii.brillien.kernel.BrillienException;
import com.vii.streamline.purity.PurityServlet;
import com.vii.streamline.services.error.StreamLineException;
import com.vii.streamline.services.json.JsonServices;
import com.vii.streamline.web.HttpServices;
import org.apache.http.HttpStatus;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

/**
 * Specialized HttpServlet to manage so called Pure-protocol, meaning to delegate BrillienCommunication calls to presences.
 */
public class PureServlet extends PurityServlet {

    /**
     * Singleton instance to be published by the webserver
     */
    protected static PureServlet servletInstance;

    static{
        servletInstance = new PureServlet();
    }

    /**
     * mediator used to communicate with presences
     */
    protected BrillienMediator                  mediator;

    /**
     * Following the singleton model, retrieves the single instance created
     */
    public static PureServlet getServletInstance() {
        return servletInstance;
    }


    {
        initMediatorContext();

        initReferences();
    }


    /**
     * Init method to set up mediator connectivity. Defined for subtypes
     */
    protected void initMediatorContext(){

    }

    /**
     * Init method to collected resources. Defined for subtypes
     */
    protected void initReferences(){

    }

    /**
     * Init method to set up mediator connectivity when the servlet is initiated by the web server
     */
    protected void setupMediatorContext( ServletConfig servletConfig ){

    }

    /**
     * Initiates mediator
     */
    protected BrillienMediator getMediator() throws ServletException {
        if( mediator == null ){
            String username = getServletConfig().getInitParameter("username");
            String password = getServletConfig().getInitParameter("password");
            if( username == null ){
                username = "dispatcher" + this.getClass().getSimpleName(); password = "dispatcher";
            }
            try {
                if( mediator == null ){
                    setupMediatorContext( getServletConfig() );
                    mediator = BrillienContext.newMediator(username, password, new LoaderComplyer() {
                        @Override
                        public ClassLoader getLoader(String reference) {
                            try {
                                return BrillienServices.getPresenceManagerByName(
                                        BrillienServices.liaison,reference).getClassLoader();
                            } catch (BrillienException e) {
                                e.printStackTrace();
                                return this.getClass().getClassLoader();
                            }
                        }
                    });
                    mediator.connect();
                }
            } catch (BrillienException e) {
                throw new ServletException( e.getMessage(), e );
            }
        }
        return mediator;
    }

    @Override
    public void init() throws ServletException {
        super.init();

        getMediator();
    }

    @Override
    protected void sendError(HttpServletRequest req, HttpServletResponse resp, int errorCode, String errorMessage, Object errorValue) {
        try {
            resp.setStatus( HttpStatus.SC_BAD_REQUEST );
            BrillienCommunication comm = mediator.newErrorCommunication(errorCode, errorMessage, errorValue);
            if( logger != null )
                logger.debug( "Sending error message:" + comm );
            HttpServices.sendJSONMessage(req, resp, comm );
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @Override
    protected Object call(String requestString) throws StreamLineException{
        PureCommunication comm = JsonServices.parseJSON( requestString, PureCommunication.class );

        try {
            return mediator.sendCommunication( comm );
        } catch (BrillienException e) {
            throw new StreamLineException( e.getErrorCode(), e.getMessage(), e.getValue() );
        }
    }

    @Override
    public void destroy() {
        try {
            if( mediator != null )
                mediator.disconnect();
        } catch (BrillienException e) { }
        BrillienContext.cleanup();
    }

}
