/*
 * Copyright (c) 2012 Imre Fazekas.
 *  All rights reserved.
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *
 *  Redistributions of source code must retain the above copyright notice, this
 *  list of conditions and the following disclaimer.
 *
 *  Redistributions in binary form must reproduce the above copyright notice,
 *  this list of conditions and the following disclaimer in the documentation
 *  and/or other materials provided with the distribution.
 *  Neither the name of the Brillien nor the names of its
 *  terms and concepts may be used to endorse or promote products derived from this
 *  software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

package com.vii.brillien.services.websocket.applications.purity;

import com.vii.streamline.services.MathServices;
import com.vii.streamline.structures.collections.InnerMap;
import org.glassfish.grizzly.http.HttpRequestPacket;
import org.glassfish.grizzly.websockets.DefaultWebSocket;
import org.glassfish.grizzly.websockets.ProtocolHandler;
import org.glassfish.grizzly.websockets.WebSocketListener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Map;

/**
 * Specialized DefaultWebSocket to extend it with entities used to manage connections and clients.
 */
public class PureWebSocket extends DefaultWebSocket {

    /**
     * Logger
     */
    protected Logger                    logger;

    /**
     * Name of the entity where this socket belongs to
     */
    protected String                    entityName;

    /**
     * Metainformation of the given Websocket. May contain additional data about the client or the connection itself
     */
    protected Map<String, Object>       properties;

    {
        logger      =   LoggerFactory.getLogger(this.getClass());
        entityName  =   MathServices.newId();
        properties  =   new InnerMap<String, Object>();
    }

    public PureWebSocket(ProtocolHandler protocolHandler, HttpRequestPacket request, WebSocketListener... listeners) {
        super(protocolHandler, request, listeners);
    }

    /**
     * Getter method for field entityName
     */
    public String getEntityName(){
        return entityName;
    }

    /**
     * Setter method for field entityName
     */
    public void setEntityName(String entityName) {
        this.entityName = entityName;
    }

    /**
     * Getter method for field properties
     */
    public Map<String, Object> getProperties() {
        return properties;
    }

    /**
     * Setter method for field properties
     */
    public void setProperties(Map<String, Object> properties) {
        this.properties = properties;
    }

    /**
     * Retrieves a metainfo identified by the given name
     */
    public <T> T getProperty(String name){
        return (T)properties.get( name );
    }

    /**
     * Tells whether metainfo identified by the given name exists
     */
    public boolean hasProperty(String name){
        return properties.containsKey( name );
    }

    /**
     * Adds a new metainfo
     */
    public void addProperty(String name, Object value){
        properties.put( name, value );
    }

    /**
     * Removes a metainfo identified by the given name
     */
    public void removeProperty(String name){
        properties.remove( name );
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (!(o instanceof PureWebSocket)) return false;

        PureWebSocket that = (PureWebSocket) o;

        if (entityName != null ? !entityName.equals(that.entityName) : that.entityName != null) return false;

        return true;
    }

}
